import numpy as np
import torch
import random
import math
import torch.nn as nn
from torch.optim import SGD


torch.backends.cudnn.benchmark=True

torch.manual_seed(1234)
torch.cuda.manual_seed_all(1234)

class OneShot_Block(nn.Module):
    """
    this net exploits the ridge loss to calculate target active feature
    """
    def __init__(self, filter_size):
        """
        args:
            filter_size - [batch, channel, height, width]
        """
        super(OneShot_Block, self).__init__()

        self.channels = filter_size[1]
        self.conv = nn.Conv2d(
            in_channels = self.channels,
            out_channels = self.channels,
            kernel_size = (3, 3),
            padding = (1, 1)
            )
        self.apply(init_weights)

    def forward(self, feature):
        return self.conv(feature)

def init_weights(module):
    if isinstance(module,nn.Conv2d):
        batch, channel, height, width = module.weight.data.shape
        module.weight.data = torch.randn(module.weight.data.shape)/(math.sqrt(channel*(height-1)/2*(width-1)/2)*1e8)
        nn.init.constant_(module.bias.data,0.0)


def oneshot_model(features, filter_sizes, device, conv51_weight, aug):
    feaure_groups_num = len(features)
    aug_templates_num = len(aug)
    max_epochs = 10
    template_num = 15

    for i in range(feaure_groups_num):
        # get Conv4-3 features
        if i==1:
            feature, filter_size = features[i], filter_sizes[i]
            reg = OneShot_Block(filter_size).to(device)
            reg.conv.weight.data = conv51_weight
            feature_size = torch.tensor(feature.shape).numpy()
            output_sigma = filter_size[-2:]* 0.1
            gauss_label = generate_gauss_label(feature_size[-2:], output_sigma).to(device)
            objective = nn.MSELoss()
            optim = SGD(reg.parameters(), lr = 1e-6, momentum=0.9, weight_decay=100)
            # fill training set
            for j in range(template_num - aug_templates_num):
                aug.append(features)
            random.shuffle(aug)
            train_reg(reg, optim, aug, objective, gauss_label, device, max_epochs)
            reg_weights = reg.conv.weight.data
    return reg_weights


def train_reg(model, optim, feature, objective, gauss_label, device, epochs = 100):
    """
    function: train the regression net and regression loss
    """
    input = feature[0][1]
    N_z, G_z, H_z, W_z = input.shape
    gauss_label_block = torch.Tensor(G_z, H_z, W_z).cuda()

    for cid, cc in enumerate(gauss_label_block):
        gauss_label_block[cid] = gauss_label * input[0][cid].max()
    for i in range(epochs):
        for ff in feature:
            realyf = ff[1]
            realyf = realyf.to(device)
            predict = model(realyf)
            predict = predict.view(1,-1)
            gauss_label = gauss_label_block.view(1,-1)
            loss = objective(predict, gauss_label)
            if hasattr(optim,'module'):
                optim.module.zero_grad()
                loss.backward()
                optim.module.step()
            else:
                optim.zero_grad()
                loss.backward()
                optim.step()


def generate_gauss_label(size, sigma, center = (0, 0), end_pad=(0, 0)):
    """
    function: generate gauss label for L2 loss
    """
    shift_x = torch.arange(-(size[1] - 1) / 2, (size[1] + 1) / 2 + end_pad[1])
    shift_y = torch.arange(-(size[0] - 1) / 2, (size[0] + 1) / 2 + end_pad[0])

    shift_y, shift_x = torch.meshgrid(shift_y, shift_x)

    alpha = 0.2
    gauss_label = torch.exp(-1*alpha*(
                        (shift_y-center[0])**2/(sigma[0]**2) +
                        (shift_x-center[1])**2/(sigma[1]**2)
                        ))

    return gauss_label

